import csv
from SimPy.Simulation import *

################################################################
# constants
################################################################
NO_JOB = -1
TIMEOUT = 1800.0 # maximal timeout for solvers

################################################################
# Instance class
################################################################

class Instance():
    def __init__(self, string):
        self.threeC = string[0]
        self.max = string[1]
        self.avg = string[2]
        self.stdv = string[3]
        self.fs = string[4]
        self.mistralclass = string[5]
        self.regft = float(string[6])
        self.ft = float(string[7])
        self.mistral = float(string[8])
        self.choco = float(string[9])
        self.abscon = float(string[10])
        self.cphydra = float(string[11])
        self.regMistral = float(string[12])
        self.regChoco = float(string[13])
        self.regAbscon = float(string[14])

    def get_time(self, solver):
        if (solver == "mistral"):
            return self.mistral
        elif (solver == "choco"):
            return self.choco
        elif (solver == "abscon"):
            return self.abscon
        elif (solver == "cphydra"):
            return self.cphydra
        else:
            print "Error in function get_time"

    def get_best_solver(self):
        if self.mistral == min(self.mistral, self.choco, self.abscon):
            return "mistral"
        elif self.choco == min(self.mistral, self.choco, self.abscon):
            return "choco"
        else:
            return "abscon"

    def get_satzilla_solver_order(self):
        mydict = {
            'mistral': self.regMistral,
            'choco': self.regChoco,
            'abscon': self.regAbscon
            }
        list = sorted([(value,key) for (key,value) in mydict.items()])
        return [key for (value,key) in list]


################################################################
# Functions that can be used to sort a list of instances
################################################################

def threec_key(x):
    if (x.threeC == "easy4all"):
        return 0
    if (x.threeC == "at_least_one_easy"):
        return 1
    if (x.threeC == "never_easy"):
        return 2

def avg_key(x):
    if (x.avg == "easy"):
        return 0
    if (x.avg == "medium"):
        return 1
    if (x.avg == "hard"):
        return 2

def max_key(x):
    if (x.max == "easy"):
        return 0
    if (x.max == "hard"):
        return 1

def stdv_key(x):
    if (x.stdv == "easy"):
        return 0
    if (x.stdv == "hard"):
        return 1

def mistral_key(x):
    if (x.mistralclass == "easy"):
        return 0
    if (x.mistralclass == "medium"):
        return 1
    if (x.mistralclass == "hard"):
        return 2

def invert(funct):
    return lambda x: funct(x)*(-1)

def fast_key(x):
    return x.ft

def regft_key(x):
    return x.regft


def compose_sort(keys):
    def compose_sort_aux(x,keys):
        for i in range(len(keys),0,-1):
            x.sort(key = keys[i-1])
    return lambda x: compose_sort_aux(x,keys)

def selection_1(job):
    return (
        job.threeC == "at_least_one_easy"
        )

def selection_2(job):
    return (
        job.stdv == "hard"
        )

def selection_3(job):
    return (
        job.threeC == "at_least_one_easy"
        and
        job.stdv == "hard"
        )



"""
def selection_4(job):
    return (
        job.threeC == "at_least_one_easy"
        and
        job.max == "hard"
        )

def selection_5(job):
    return (
        job.stdv == "hard"
        and
        job.max == "hard"
        )

def selection_6(job):
    return (
        not (job.threeC == "never_easy")
        )
"""


################################################################
# Classes used for the simulation
################################################################

class Interruptor(Process):
    """
    Used to interrupt a list of clients
    """
    def stop(self,list):
        #print "Interrupting clients %s" % ([i.name for i in list])
        for i in list:
            self.interrupt(i)

class Client(Process):
    """
    Client implementation supporting interruption
    """

    def __init__(self,name,server):
        Process.__init__(self,name=name)
        self.server = server

    def start(self):
        (job,solver,user_timeout) = self.server.getjob(self)
        running_time = 0.0    
        while job != NO_JOB:
            #print "todel client %s received job %s at time %s" % (self.name, (job,solver,user_timeout), now())
            time = now()
            timeout = self.server.instances[job].get_time(solver)
            #print timeout
            timeout = min(timeout,user_timeout)
            self.server.solver_calls[solver] += 1
            yield hold,self,timeout
            running_time = now() - time
            if not(self.interrupted()):
                if timeout != user_timeout:
                    self.server.finished(self,job)
                else:
                    self.server.not_finished(self,job)
            else:
                #print "client %s has been interrupted" %(self.name)
                self.server.interrupts_num += 1
            (job,solver,user_timeout) = self.server.getjob(self)
        self.server.stop(self,running_time)


class Schedule(object):
    """
    Basic implementation for the optimal scheduler.
    Inheredit from this class defining functions create_jobs,
    finished, not_finished and getjob to create new scheduling
    strategies.
    Note that a super call to finished and not_finished is needed
    in the derived classes
    """

    def __init__(self,file_name,*args):
        """
        file_name is the cvs file path to use
        args is a list containing the sorting function
        """

        sorting_function = args[0]

        # stores the solving times for every job
        self.solving_times = {}
        # stores the finishing times of every processor
        self.finishing_times = {}
        # stores the running times of every processor
        self.running_times = {}
        # all the instances to solve
        self.instances = []
        #number of interrupts
        self.interrupts_num = 0

        self.solver_calls = {}
        self.solver_calls["mistral"] = 0
        self.solver_calls["choco"] = 0
        self.solver_calls["abscon"] = 0
        self.solver_calls["cphydra"] = 0

        reader = csv.reader(open(file_name, "rb"))
        row_number = 0
        for row in reader:
            if row_number != 0:
                self.instances.append(Instance(row))
            else:
                row_number += 1
        sorting_function(self.instances)

        

    def stop(self,client,time):
        self.finishing_times[client.name] = now()
        self.running_times[client.name] = time
        #print "todel client %s finished at %s" % (client.name, self.finishing_times[client.name])

    def create_jobs(self): # used to create the jobs
        self.instances.reverse()
        self.jobs = map(
            lambda x: (x,self.instances[x].get_best_solver(),TIMEOUT),
            range(len(self.instances))
            )

    def getjob(self,client):
        #print "getjob request: client %s" % (client.name)
        if len(self.jobs) == 0:
            #print "no more jobs available at time %s" % (now())
            return (NO_JOB,"",TIMEOUT)
        else:
            job = self.jobs.pop()
            #print "sending the job %s to client %s"% (job[0],client.name)
            return job

    def finished(self,client,job): #job finished
        self.solving_times[job] = now()
        #print "job %s finished by client %s at time %s" % (job,client.name,now())

    def not_finished(self,client,job): #job not finished
        pass
        #print "job %s was not finished by client %s at time %s" % (job,client.name,now())

