# Code that cpHydra uses to compute the solvers to use
# Taken from the Numberjack project

"""
CpHydra does not pay attention to the execution order of the solvers.
In our case since the solving time makes a difference we chose to
use the solver form the fastest to the slowest in average
The order is therefore Mistral, Choco, Abscon
"""

import csv
import Mistral
from Numberjack import *
from Constants import *


epsilon = 0.0001

"""
get_cpHydra_time gets an instance (a row of the csv file),
the dataset of training cases and computes the solving time
required by CpHydra to solve that instance
"""

def get_cpHydra_time(test_row,cases):
    target = Case(test_row)
    solver_map = get_cpHydra_mapping(target,cases)
    real_time = {}
    real_time['mistral'] = float(test_row[-7])
    real_time['choco'] = float(test_row[-8])
    real_time['abscon'] = float(test_row[-9])

    time = 0.0
    solver = 'mistral'
    if real_time[solver] <= solver_map[solver]:
        return real_time[solver]
    else:
        time += solver_map[solver]
        del(solver_map[solver])
        solver = 'choco'
        if real_time[solver] <= solver_map[solver]:
            return time + real_time[solver]
        else:
            time += solver_map[solver]
            del(solver_map[solver])
            solver = 'abscon'
            if real_time[solver] <= solver_map[solver]:
                return time + real_time[solver]
            else:
                return time + solver_map[solver]



def get_cpHydra_mapping(target,cases):
    return schedule(
        cases.get_similar(target),
        ['mistral','choco','abscon'],
        1800)

def schedule(cases, solvers, total_time):
    def weight(case):
        d = case[0]
        if d >= 1000: return 1
        elif d <= 0.01: return 100000
        else: return int(1000/d)

    model = Model()

    allocated_times = VarArray(len(solvers), 0, total_time)
    prob_solved = VarArray(len(cases))

    model.add(Sum(allocated_times) == total_time)

    for prob, case in zip(prob_solved, cases):
        expected_times = [case[1].expected_solve_time(solver) for solver in solvers]
        model.add(prob == (Sum([stime >= int(etime+1) for stime, etime in
                                zip(allocated_times, expected_times)]) >= 1))

    model.add(Maximise(Sum(prob_solved, [weight(case) for case in cases])))

    solver = Mistral.Solver(model, allocated_times)
    solver.setHeuristic('MinDomain','RandomSplit',3)

    solver.solveAndRestart()

    return dict(zip(solvers, map(value, allocated_times)))

class Case:
    def __init__(self, name='no_name', f={}, b={}):
        self.name = name
        self.features = f
        self.beliefs = b

    def __init__(self, vector):
        self.name = 'no_name'
        self.features = dict(zip(range(42), map(float,vector[:42])))
        self.beliefs = {
            'mistral':float(vector[-7]),
            'choco':float(vector[-8]),
            'abscon':float(vector[-9])
            }

    def compare(self, case):
        common_keys = (frozenset(self.features.keys()) &
                       frozenset(case.features.keys()))

        distance = 0
        for k in common_keys:
            d = (self.features[k] - case.features[k])
            distance += (d*d)

        return 1+distance

    def expected_solve_time(self,solver,default=1800.0):
        if self.beliefs.has_key(solver): return self.beliefs[solver]
        else: return default

    def __str__(self):
        print "features "
        print self.features
        print "beliefs "
        print self.beliefs


class Casebase(list):
    '''
    This will store the casebase
    '''
    def __init__(self, cases=[]):
        list.__init__(self, cases)

    def __str__(self):
        return '\n'.join(map(str, self))

    def get_probabilities(self, case, solver, k=10):
        global epsilon
        W = 0
        data = {}
        cases = self.get_similar(case, k)
        for d,c in cases:
            if d>1+epsilon:
                t = c.expected_solve_time(solver)
                if data.has_key(t): data[t] += 1/d
                else: data[t] = 1/d
                W += 1/d
        runtimes = data.keys()
        runtimes.sort()
        proba = [(0,0)]
        p = 0
        for r in runtimes:
            p += data[r]
            proba.append((r, p/W))
        return proba

    def get_similar(self, case, k=10):
        comparisons = [(case.compare(c), c) for c in self]
        comparisons.sort()
        #return [case for (_, case) in comparisons[:k]]
        return comparisons[:k]

    def add(self,case):
        '''
        This method adds a case to the base,
        or updates the current case if a similar exists
        '''
        #(distance,most_similar) = min([(case.compare(c), c) for c in self])
        #if distance <= epsilon: # we assume that it is the same instance
        #    for k in case.keys(): most_similar[k] = case[k]
        #else:
        self.append(case)

    def addFromFile(self,file):
        reader = csv.reader(open(file, "rb"), delimiter=';')
        row_number = 0
        for row in reader:
            if row_number != 0:
                self.add(Case(row))
            else:
                row_number +=1